#!/usr/bin/env python
"""PyTekSwift.DataLoader.

DESCRIPTION OF dataloader MODULE:

Utility functions to read from the swift database files, lookup swiftcodes,
and lookup by countries.
"""

from yaml import load  # ,dump
import os

try:
    from yaml import CLoader as Loader  # CDumper as Dumper
except ImportError:
    from yaml import Loader  # , Dumper

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


class InvalidCountryCode(Exception):
    """Exception Class for Invalid alpha_2,alpha_3 country codes."""

    pass


class InvalidSwiftCode(Exception):
    """InvalidSwiftCode Exception."""

    pass


def build_path(countrycode: str) -> str:
    """Build Path to YAML File."""
    countrycode = countrycode.lower()
    if not len(countrycode) >= 2:
        raise InvalidCountryCode("CountryCode Too Short <2")
    path = os.path.join(CURRENT_DIR, f"data/{countrycode[:2]}.yml")
    if not os.path.exists(path):
        raise InvalidCountryCode
    return path


def load_country_data(countrycode):
    """Return a Python Object from the loaded yaml countrycode data file."""
    path = build_path(countrycode)
    with open(path, "r") as fl:
        document = load(fl, Loader=Loader)
    return document


def isvalid_swiftcode(code):
    """Is swiftcode valid, 8 chars or 11 chars?."""
    if len(code) == 8 or len(code) == 11:
        return True
    else:
        return False


def countrycode_from_swiftcode(swiftcode: str):
    """Swift Code Structure.

    The previous edition is ISO 9362:2009 (dated 2009-10-01).
    The SWIFT code is 8 or 11 characters, made up of:

    4 letters: institution code or bank code.
    2 letters: ISO 3166-1 alpha-2 country code (exceptionally,
    SWIFT has assigned the code XK to Republic of Kosovo,
    which does not have an ISO 3166-1 country code)
    2 letters or digits: location code
        if the second character is "0",
        then it is typically a test BIC as opposed to a
        BIC used on the live network.  if the second character is "1",
        then it denotes a passive participant in the SWIFT network
        if the second character is "2",
        then it typically indicates a reverse billing BIC,
        where the recipient pays for the message as opposed to the more usual
        mode whereby the sender pays for the message.
    3 letters or digits: branch code, optional ('XXX' for primary office)

    Where an eight digit code is given,
    it may be assumed that it refers to the primary office.

    """
    if not isvalid_swiftcode(swiftcode):
        raise InvalidSwiftCode(f"Invalid Swift Code {swiftcode}")
    return f"{swiftcode[4:6]}"


def lookup_swiftcode(swiftcode: str):
    """Get Details of provided swiftcode or None."""
    data = load_country_data(countrycode_from_swiftcode(swiftcode))
    swiftkey = swiftcode.upper()
    return data.get(swiftkey)



"""tekswift package.

Defines:
    build_path,
    countrycode_from_swiftcode,
    load_country_data,
    isvalid_swiftcode,
    InvalidCountryCode,
    InvalidSwiftCode,
"""
# import
from .dataloader import (
    InvalidCountryCode,
    InvalidSwiftCode,
    build_path,
    countrycode_from_swiftcode,
    isvalid_swiftcode,
    load_country_data,
    lookup_swiftcode,
)


__all__ = [
    "InvalidCountryCode",
    "InvalidSwiftCode",
    "build_path",
    "countrycode_from_swiftcode",
    "isvalid_swiftcode",
    "load_country_data",
    "lookup_swiftcode",
]

__version__ = "0.1.2"

__author__ = "Tralah M Brian <https://github.com/TralahM>"



"""tekswift cli (command line interface)."""
import argparse
import yaml
from tekswift import dataloader
from tekswift.utils import transform_country_data, bin_to_swifts


def bin2swift_lookup(args: argparse.Namespace):
    """Return the Swift codes for the matching issuer."""
    swifts, bin_data, pmap = bin_to_swifts(
        args.bin_code,
        args.threshold,
        fuzz=True,
    )
    bin_data["swift_data"] = swifts
    print(yaml.dump(bin_data))
    if args.verbose:
        # [print(f"{k}: {b}") for k, b in pmap.items()]
        print(pmap)
    # print(yaml.dump(swifts))
    return


def swift_lookup(args: argparse.Namespace):
    """Lookup the given SWIFTCODE in the database."""
    info = dataloader.lookup_swiftcode(args.swiftcode)
    print(yaml.dump(info))
    return


def country_lookup(args: argparse.Namespace):
    """Return details of all banks for the given country code."""
    country_data = dataloader.load_country_data(args.country_code)
    print(yaml.dump(transform_country_data(country_data)))
    return


def usage(args, **kwargs):
    """Handle Main repodoc Entrypoint without subcommands."""
    if args.bash_completion:
        import shtab

        print(shtab.complete(args.parser, shell="bash"))
        return
    if args.zsh_completion:
        import shtab

        print(shtab.complete(args.parser, shell="zsh"))
        return


def get_main_parser():
    """Return main argparse.ArgumentParser."""
    parser = argparse.ArgumentParser()
    parser.set_defaults(func=usage)
    parser.add_argument(
        "-bc",
        "--bash-completion",
        help="Print Bash Completion Script.",
        action="store_true",
        dest="bash_completion",
    )
    parser.add_argument(
        "-zc",
        "--zsh-completion",
        help="Print ZSH Completion Script.",
        action="store_true",
        dest="zsh_completion",
    )
    subparsers = parser.add_subparsers(title="subcommands")
    lookup_sp = subparsers.add_parser(
        "lookup",
        aliases=["l", "lk", "get"],
        help=swift_lookup.__doc__,
    )
    lookup_sp.set_defaults(func=swift_lookup)
    lookup_sp.add_argument(
        "swiftcode", help="SWIFTCODE [8 or 11 alphanumeric].")
    country_sp = subparsers.add_parser(
        "country",
        aliases=["cn"],
        help=country_lookup.__doc__,
    )
    country_sp.set_defaults(func=country_lookup)
    country_sp.add_argument(
        "country_code", help="ISO-2 or ISO-3 Country Code.")
    bin_sp = subparsers.add_parser(
        "bin",
        aliases=["bn"],
        help=bin2swift_lookup.__doc__,
    )
    bin_sp.set_defaults(func=bin2swift_lookup)
    bin_sp.add_argument(
        "-p",
        "--probability-threshold",
        action="store",
        default=0.8,
        dest="threshold",
        help="threshold to use as a filter bettween [0,1].",
        type=float,
    )
    bin_sp.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        dest="verbose",
        help="be verbose.",
    )
    bin_sp.add_argument("bin_code", help="Bank Indetification Code.")
    return parser


def main():
    """Run main cli entry point."""
    parser = get_main_parser()
    args = parser.parse_args()
    args.func(args)
    return



"""Utilities for deriving functionality from pytekswift and pycctek."""
import string
import collections
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import CountVectorizer
from fuzzywuzzy import fuzz

try:
    from nltk.corpus import stopwords
except LookupError:
    import nltk

    nltk.download("stopwords")

import cctek
from tekswift import dataloader

stopwords = stopwords.words("english")


def fuzz_prob(issuer, institution):
    """Return Fuzz wuzz simple ratio divided by 100."""
    return fuzz.ratio(issuer, institution) / 100


def transform_country_data(country_data):
    """Return a new flattened representation of the swiftcode_data."""
    new_data = []
    for key, value in country_data.items():
        new_transform = collections.defaultdict()
        new_transform["swiftcode"] = key
        new_transform.update(value)
        new_data.append(new_transform)
    return new_data


def clean_string(text):
    """Return cleaned string."""
    text = "".join([word for word in text if word not in string.punctuation])
    text = text.lower()
    text = "".join([word for word in text.split() if word not in stopwords])
    return text


def cosine_sim_vectors(vec1, vec2):
    """Calculate the cosine similarity of 2 vectors.

    Note `cosine_similarity()` expects 2D arrays and the input vectors are 1D
    vectors so we need to reshape them.
    """
    vec1 = vec1.reshape(1, -1)
    vec2 = vec2.reshape(1, -1)
    return cosine_similarity(vec1, vec2)[0][0]


def vectorize(cleaned_strings: list):
    """Return vectors of CountVectorizer.

    Where vectors[0] is first string,and vectors[1] is the second string.
    """
    vectorizer = CountVectorizer().fit_transform(cleaned_strings)
    vectors = vectorizer.toarray()
    return vectors


def issuer_institution_similarity(issuer: str, institution: str) -> float:
    """Return prob[0,1] that issuer and institution strings are similar."""
    cleaned = list(map(clean_string, [issuer, institution]))
    vectors = vectorize(cleaned)
    probability = cosine_sim_vectors(vectors[0], vectors[1])
    return probability


def get_institutions_from_swift(country_code):
    """Return a list of institutions and swiftcodes from country code."""
    return transform_country_data(dataloader.load_country_data(country_code))


def issuer_institutions_probmap(issuer, institutions, fuzz=False):
    """Return a map of swiftcode to probabilty"""
    probmap = collections.defaultdict()
    for ins in institutions:
        if fuzz:
            probmap[ins.get("swiftcode")] = fuzz_prob(
                issuer, ins.get("institution"))
        else:
            probmap[ins.get("swiftcode")] = issuer_institution_similarity(
                issuer, ins.get("institution")
            )
    return probmap


def filter_probmap(probmap, probability):
    """Return a filtered list of swiftcodes."""
    return [k for k, v in probmap.items() if v >= probability]


def bin_to_swifts(bin_no, p=0.8, fuzz=True):
    """Bank Identification Number to Swift matches."""
    bin_data = cctek.bin_checker(bin_no)
    issuer = bin_data.get("issuer")
    country_code = bin_data.get("country").get("alpha_2")
    institutions = get_institutions_from_swift(country_code)
    probmap = issuer_institutions_probmap(issuer, institutions, fuzz=fuzz)
    swifts = {c: dataloader.lookup_swiftcode(
        c) for c in filter_probmap(probmap, p)}
    return swifts, bin_data, probmap
